import sqlalchemy as sa
from sqlalchemy import Table, Column, ForeignKey
from sqlalchemy import Integer, String, Boolean, JSON, Enum, DateTime
from sqlalchemy.orm import relationship, validates
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.ext.hybrid import hybrid_property
import datetime
import re
import copy
import types
from . import _utils as utils
from . import _timeseries as ts
from . import _storage as storage
from . import _connection as conn


Base = declarative_base()
valid_name = re.compile(r"^[a-zA-Z0-9\.#_-]+$")
partitions = Enum("year", "date", name="partition")


class FeatureStoreMixin(object):
    name = Column(String(128), primary_key=True, nullable=False)
    description = Column(String, default="")
    meta = Column(JSON, default={})
    version = Column(Integer, default=1, nullable=False)

    @validates("name")
    def _validate_name(self, key, value):
        if not valid_name.match(value):
            raise ValueError(f"Invalid name {value}")
        return value

    def as_dict(self):
        return {
            k: v if utils.is_jsonable(v) else str(v)
            for k, v in self.__dict__.items()
            if k[0] != "_"
        }

    def bump_version(self):
        if self.version:
            self.version += 1
        else:
            self.version = 1

    def update_from_dict(self, payload):
        if not payload:
            return
        if "name" in payload and self.name:
            raise ValueError(
                f"Cannot change name of {self.__class__.__name__}: use clone instead"
            )
        if "namespace" in payload and self.namespace:
            raise ValueError(
                f"Cannot change namespace of {self.__class__.__name__}: use clone instead"
            )
        for key, value in payload.items():
            if key == "meta" or key == "metadata":
                # See https://amercader.net/blog/beware-of-json-fields-in-sqlalchemy/
                self.meta = copy.deepcopy(self.meta)
                # Merge old metadata with new
                if self.meta:
                    self.meta.update(value)
                else:
                    self.meta = value
                # Remove any keys that no longer have values
                self.meta = {k: v for k, v in self.meta.items() if v is not None}
            else:
                # Update fields
                setattr(self, key, value) if hasattr(self, key) else None
        self.bump_version()


class ByteHubVersion(Base):
    __tablename__ = "bytehub_version"

    id = Column(Integer, primary_key=True)
    timestamp = Column(DateTime, default=datetime.datetime.utcnow)
    version = Column(String(128), nullable=False)


class Namespace(Base, FeatureStoreMixin):
    __tablename__ = "namespace"

    url = Column(String, nullable=False, unique=True)
    storage_options = Column(JSON, nullable=False, default={})
    backend = Column(String, nullable=True, default="pandas")

    @hybrid_property
    def namespace(self):
        return self.name

    @namespace.setter
    def namespace(self, value):
        self.name = value

    @namespace.expression
    def namespace(cls):
        return cls.name

    def _backend(self):
        # Check backend is available and get it
        backend = "pandas" if not self.backend else self.backend.lower()
        if self.backend in storage.available_backends:
            return storage.available_backends[backend](
                url=self.url, storage_options=self.storage_options
            )
        else:
            raise RuntimeError(
                f"{backend} storage backend is not available: make sure and dependencies are installed"
            )

    def clean(self):
        # Check for unused data and remove it
        store = self._backend()
        active_feature_names = [f.name for f in self.features]
        feature_data = store.ls()
        for feature in feature_data:
            if feature not in active_feature_names:
                # Redundant data... delete it
                store.delete(feature)


class Feature(Base, FeatureStoreMixin):
    __tablename__ = "feature"

    namespace = Column(String(128), ForeignKey("namespace.name"), primary_key=True)
    namespace_object = relationship("Namespace", backref="features")

    partition = Column(partitions, default="date", nullable=False)
    serialized = Column(Boolean, default=False, nullable=False)
    transform = Column(JSON, nullable=True)

    @hybrid_property
    def full_name(self):
        return f"{self.namespace}/{self.name}"

    @validates("serialized")
    def validate_serialized(self, key, value):
        if self.serialized is not None and value != self.serialized:
            raise ValueError("Cannot change serialized setting on existing feature")
        return value

    @validates("transform")
    def validate_transform(self, key, value):
        if not value:
            return value
        if isinstance(value.get("function"), str):
            # Function already serialized, no conversion required
            func = value.get("function")
        elif isinstance(value.get("function"), types.FunctionType):
            func = utils.serialize(value["function"])
        else:
            raise ValueError(
                "Transform must be a Python function, accepting a single dataframe input"
            )
        assert "function" in value.keys(), "Transform must have a function defined"
        assert "args" in value.keys(), "Transform must have arguments defined"
        # Convert function to base64/cloudpickle format
        return {
            "format": "cloudpickle",
            "function": func,
            "args": value["args"],
        }

    @classmethod
    def clone_from(cls, other, namespace, name):
        if not isinstance(other, cls):
            raise ValueError(f"Must clone from another {cls.__name__}")
        clone = cls()
        # Build new Feature with same settings as old
        clone.namespace = namespace
        clone.name = name
        payload = other.as_dict()
        payload.pop("namespace")
        payload.pop("name")
        payload.pop("version")
        clone.update_from_dict(payload)
        return clone

    def save(self, df):
        store = self.namespace_object._backend()
        store.save(self.name, df, partition=self.partition, serialized=self.serialized)

    def load_transform(
        self, from_date, to_date, freq, time_travel, last=False, callers=[]
    ):
        # Get the SQLAlchemy session for this feature
        session = sa.inspect(self).session
        if not session:
            raise RuntimeError(f"{self.name} is not bound to an SQLAlchemy session")
        # Check for recursive transforms
        if self.full_name in callers:
            raise RuntimeError(
                f"Recursive feature transform detected on {self.full_name}"
            )
        # Load the transform function
        func = utils.deserialize(self.transform["function"])
        # Load the features to transform
        dfs = []
        # Load each requested feature
        for f in self.transform["args"]:
            namespace, name = f.split("/")[0], "/".join(f.split("/")[1:])
            feature = (
                session.query(Feature)
                .filter_by(name=name, namespace=namespace)
                .one_or_none()
            )
            if not feature:
                raise ValueError(f"No feature named {name} exists in {namespace}")
            # Load individual feature
            df = feature.load(
                from_date=from_date,
                to_date=to_date,
                freq=freq,
                time_travel=time_travel,
                last=last,
                callers=[*callers, self.full_name],
            )
            dfs.append(df.rename(columns={"value": f"{namespace}/{name}"}))
        # Merge features into a single dataframe
        dfs = ts.concat(dfs)
        # Make sure columns are in the same order as args
        dfs = dfs[self.transform["args"]]
        # Apply transform function
        transformed = ts.transform(dfs, func)
        return transformed

    def load(
        self,
        from_date=None,
        to_date=None,
        freq=None,
        time_travel=None,
        last=False,
        callers=[],
    ):
        # Does this feature need to be transformed?
        if self.transform:
            return self.load_transform(
                from_date=from_date,
                to_date=to_date,
                freq=freq,
                time_travel=time_travel,
                last=last,
                callers=callers,
            )
        # Get storage
        store = self.namespace_object._backend()
        # Restrict which partitions are loaded when getting last value
        if last:
            from_date = store.last(self.name)
            to_date = None
        # Load dataframe
        return store.load(
            self.name,
            from_date=from_date,
            to_date=to_date,
            freq=freq,
            time_travel=time_travel,
            serialized=self.serialized,
        )

    def last(self):
        # Fetch last feature value
        df = self.load(last=True)
        result = df.tail(1)
        if result.empty:
            return None
        else:
            return result["value"].iloc[0]

    def delete_data(self):
        # Deletes all of the data on this feature
        store = self.namespace_object._backend()
        store.delete(self.name)

    def import_data_from(self, other):
        # Copy data over from another feature
        if not isinstance(other, self.__class__):
            raise ValueError(f"Must clone from another {cls.__name__}")
        # Get location of other feature to copy from
        store_from = other.namespace_object._backend()
        store_to = self.namespace_object._backend()
        # Copy data to new location
        store_from.copy(other.name, self.name, store_to)



from . import _connection as conn
from . import _model as model
from . import _version as version
from alembic.operations import Operations
from alembic.migration import MigrationContext
from sqlalchemy import String, Column
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import sessionmaker
from packaging.version import parse


def upgrade(engine):
    session_maker = sessionmaker(bind=engine)
    session = session_maker()

    r = (
        session.query(model.ByteHubVersion)
        .order_by(model.ByteHubVersion.timestamp.desc())
        .first()
    )
    if not r:
        current_version = "0"
    else:
        current_version = r.version

    if parse(current_version) >= parse(version.__version__):
        # Up-to-date: nothing to do
        session.close()
        return

    if parse(current_version) < parse(version.__version__):
        print(f"Upgrading ByteHub database schema...")
        with engine.connect() as connection:
            context = MigrationContext.configure(connection)
            op = Operations(context)
            try:
                op.add_column("namespace", Column("backend", String(128)))
            except OperationalError:
                pass

    # Add version number
    obj = model.ByteHubVersion()
    obj.version = version.__version__
    session.add(obj)
    session.commit()

    session.close()



from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from contextlib import contextmanager
from pandas.io import json


def connect(conn, connect_args={}):
    engine = create_engine(conn, json_serializer=json.dumps, connect_args=connect_args)
    return engine, sessionmaker(bind=engine)


@contextmanager
def session_scope(session_maker):
    """Provide a transactional scope around a series of operations."""
    session = session_maker()
    try:
        yield session
        session.commit()
    except:
        session.rollback()
        raise
    finally:
        session.close()



import pandas as pd
import functools

try:
    # Allow for a minimal install with no dask/pyarrow
    from dask import dataframe as dd
except ImportError:
    pass


def concat(dfs):
    """Concat dataframes for multiple features."""
    if len(dfs) == 0 or isinstance(dfs[0], pd.DataFrame):
        # Use Pandas concat
        return pd.concat(dfs, join="outer", axis=1).ffill()
    elif isinstance(dfs[0], dd.DataFrame):
        # Use Dask concat
        dfs = functools.reduce(
            lambda left, right: dd.merge(
                left, right, left_index=True, right_index=True, how="outer"
            ),
            [df for df in dfs],
        )
        return dfs.ffill()
    else:
        raise NotImplementedError(f"Unrecognised dataframe type: {type(dfs[0])}")


def transform(df, func):
    """Transform dataframe using function."""
    transformed = func(df)
    # Make sure output has a single column named 'value'
    if isinstance(transformed, pd.Series) or isinstance(transformed, dd.Series):
        transformed = transformed.to_frame("value")
    if isinstance(df, pd.DataFrame) and not isinstance(transformed, pd.DataFrame):
        raise RuntimeError(
            f"Transforms in this namespace should return Pandas dataframes or series"
        )
    if isinstance(df, dd.DataFrame) and not isinstance(transformed, dd.DataFrame):
        raise RuntimeError(
            f"Transforms in this namespace should return Dask dataframes or series"
        )
    if len(transformed.columns) != 1:
        raise RuntimeError(
            f"Transform function should return a dataframe with a datetime index and single value column"
        )
    transformed.columns = ["value"]
    return transformed



import importlib
import os


available_backends = {}


def _import_backend(name):
    try:
        mod = importlib.import_module("." + name, __name__)
        available_backends[name] = mod.Store
    except ImportError:
        pass


for module in os.listdir(os.path.dirname(__file__)):
    if module.startswith("_") or module[-3:] != ".py":
        continue
    _import_backend(module[:-3])
del module



"""
ByteHub provides an easy-to-use Feature Store, optimised for time-series data.
It requires no complex infrastructure setup, and can be run locally, connected to
a remote database/file storage, or run in a cloud-hosted mode.
ByteHub uses Dask for data storage, allowing it to be scaled to large datasets and
cluster compute environments.

Example usage for a local SQLite feature store:

    import bytehub as bh
    fs = bh.FeatureStore()

Remote feature stores can be accessed using a SQLAlchemy connection string, e.g.:

    fs = bh.FeatureStore('postgresql+psycopg2://user:pass@host:port/bytehub')

Cloud-hosted feature stores can be accessed via a REST API endpoint:

    fs = bh.FeatureStore('https://api.bytehub.ai/')

or simply use the following to access ByteHub's cloud service:

    fs = bh.CloudFeatureStore()

See [https://docs.bytehub.ai](https://docs.bytehub.ai) for examples and tutorials.

"""
from .core import CoreFeatureStore
from .cloud import CloudFeatureStore
from ._version import __version__


def FeatureStore(connection_string="sqlite:///bytehub.db", **kwargs):
    """Factory method to create Feature Store objects.

    Args:
        connection_string (str): SQLAlchemy connection string for database
            containing feature store metadata (defaults to local sqlite file)
            or an HTTPS endpoint to a cloud-hosted feature store.
        **kwargs: Additional options to be passed to the Feature Store constructor.

    Returns:
        Union[CoreFeatureStore, CloudFeatureStore]: Feature Store object.
    """
    if connection_string.startswith("http"):
        # Connect to cloud-hosted feature store
        return CloudFeatureStore(connection_string=connection_string, **kwargs)
    else:
        # Direct connection to database using CoreFeatureStore
        return CoreFeatureStore(connection_string=connection_string, **kwargs)



import sqlalchemy as sa
import pandas as pd
import posixpath
import json
from sqlalchemy.sql import text
from . import _connection as conn
from ._base import BaseFeatureStore
from . import _timeseries as ts
from . import _model as model
from . import _upgrade as upgrade
from .exceptions import *


class CoreFeatureStore(BaseFeatureStore):
    """**Core Feature Store**

    Connects directly to a SQLAlchemy-compatible database.

    When using specifying features
    for `create_feature`, `update_feature`, etc., use either:

    * `namespace` and `name` as arguments; or
    * specify `name` in the format `"my-namespace/my-feature"`.
    """

    def __init__(
        self,
        connection_string="sqlite:///bytehub.db",
        connect_args={},
    ):
        """
        Args:
            connection_string (str): SQLAlchemy connection string for database
                containing feature store metadata - defaults to local sqlite file.
            connect_args (dict, optional): dictionary of [connection arguments](https://docs.sqlalchemy.org/en/14/core/engines.html#sqlalchemy.create_engine.params.connect_args)
                to pass to SQLAlchemy.
        """
        self.engine, self.session_maker = conn.connect(
            connection_string, connect_args=connect_args
        )
        model.Base.metadata.create_all(self.engine)
        # Upgrade the database schema if required
        upgrade.upgrade(self.engine)

    def _list(self, cls, namespace=None, name=None, regex=None, friendly=True):
        namespace, name = self.__class__._split_name(namespace, name)
        with conn.session_scope(self.session_maker) as session:
            r = session.query(cls)
            # Filter by namespace
            if namespace:
                r = r.filter_by(namespace=namespace)
            # Filter by matching name
            if name:
                r = r.filter_by(name=name)
            objects = r.all()
            df = pd.DataFrame([obj.as_dict() for obj in objects])
            if df.empty:
                return pd.DataFrame()
            # Filter by regex search on name
            if regex:
                df = df[df.name.str.contains(regex)]
            # List transforms as simple true/false
            if "transform" in df.columns and friendly:
                df = df.assign(transform=df["transform"].apply(lambda x: x is not None))
            # Sort the columns
            column_order = ["namespace", "name", "version", "description", "meta"]
            column_order = [c for c in column_order if c in df.columns]
            df = df[[*column_order, *df.columns.difference(column_order)]]
            return df

    def _delete(self, cls, namespace=None, name=None, delete_data=False):
        namespace, name = self.__class__._split_name(namespace, name)
        with conn.session_scope(self.session_maker) as session:
            r = session.query(cls)
            if namespace:
                r = r.filter_by(namespace=namespace)
            if name:
                r = r.filter_by(name=name)
            obj = r.one_or_none()
            if not obj:
                raise MissingFeatureException(
                    f"No existing {cls.__name__} named {name} in {namespace}"
                )
            if hasattr(obj, "delete_data") and delete_data:
                obj.delete_data()
            session.delete(obj)

    def _update(self, cls, namespace=None, name=None, payload={}):
        namespace, name = self.__class__._split_name(namespace, name)
        with conn.session_scope(self.session_maker) as session:
            r = session.query(cls)
            if namespace:
                r = r.filter_by(namespace=namespace)
            if name:
                r = r.filter_by(name=name)
            obj = r.one_or_none()
            if not obj:
                raise MissingFeatureException(
                    f"No existing {cls.__name__} named {name} in {namespace}"
                )
            # Apply updates from payload
            obj.update_from_dict(payload)

    def _create(self, cls, namespace=None, name=None, payload={}):
        if cls is model.Namespace:
            payload.update({"name": name})
        else:
            namespace, name = self.__class__._split_name(namespace, name)
            if not self._exists(model.Namespace, namespace=namespace):
                raise MissingFeatureException(f"{namespace} namespace does not exist")
            payload.update({"name": name, "namespace": namespace})
        with conn.session_scope(self.session_maker) as session:
            obj = cls()
            obj.update_from_dict(payload)
            session.add(obj)

    def _exists(self, cls, namespace=None, name=None):
        ls = self._list(cls, namespace=namespace, name=name)
        return not ls.empty

    def list_namespaces(self, **kwargs):
        self.__class__._validate_kwargs(kwargs, ["name", "namespace", "regex"])
        return self._list(
            model.Namespace,
            name=kwargs.get("name", kwargs.get("namespace")),
            regex=kwargs.get("regex"),
        )

    def create_namespace(self, name, **kwargs):
        self.__class__._validate_kwargs(
            kwargs,
            valid=["description", "url", "storage_options", "backend", "meta"],
            mandatory=["url"],
        )
        self._create(model.Namespace, name=name, payload=kwargs)

    def update_namespace(self, name, **kwargs):
        self.__class__._validate_kwargs(
            kwargs,
            valid=["description", "storage_options", "meta"],
        )
        self._update(model.Namespace, name=name, payload=kwargs)

    def delete_namespace(self, name):
        if not self.list_features(namespace=name).empty:
            raise FeatureStoreException(
                f"{name} still contains features: these must be deleted first"
            )
        self._delete(model.Namespace, name=name)

    def clean_namespace(self, name):
        with conn.session_scope(self.session_maker) as session:
            r = session.query(model.Namespace)
            r = r.filter_by(name=name)
            namespace = r.one_or_none()
            if not namespace:
                raise MissingFeatureException(f"No existing Namespace named {name}")
            namespace.clean()

    def list_features(self, **kwargs):
        self.__class__._validate_kwargs(
            kwargs, valid=["name", "namespace", "regex", "friendly"]
        )
        return self._list(
            model.Feature,
            namespace=kwargs.get("namespace"),
            name=kwargs.get("name"),
            regex=kwargs.get("regex"),
            friendly=kwargs.get("friendly", True),
        )

    def create_feature(self, name, namespace=None, **kwargs):
        self.__class__._validate_kwargs(
            kwargs,
            valid=["description", "meta", "partition", "serialized", "transform"],
            mandatory=[],
        )
        self._create(model.Feature, namespace=namespace, name=name, payload=kwargs)

    def clone_feature(self, name, namespace=None, **kwargs):
        self.__class__._validate_kwargs(
            kwargs,
            valid=["from_namespace", "from_name"],
            mandatory=["from_name"],
        )
        from_namespace, from_name = self.__class__._split_name(
            kwargs.get("from_namespace"), kwargs.get("from_name")
        )
        to_namespace, to_name = self.__class__._split_name(namespace, name)
        if not self._exists(model.Namespace, namespace=from_namespace):
            raise MissingFeatureException(f"{from_namespace} namespace does not exist")
        if not self._exists(model.Namespace, namespace=to_namespace):
            raise MissingFeatureException(f"{to_namespace} namespace does not exist")
        with conn.session_scope(self.session_maker) as session:
            # Get the existing feature
            r = session.query(model.Feature)
            r = r.filter_by(namespace=from_namespace, name=from_name)
            feature = r.one_or_none()
            if not feature:
                raise MissingFeatureException(
                    f"No existing Feature named {from_name} in {from_namespace}"
                )
            # Create the new feature
            new_feature = model.Feature.clone_from(feature, to_namespace, to_name)
            session.add(new_feature)
            # Copy data to new feature, if this raises exception will rollback
            if not new_feature.transform:
                new_feature.import_data_from(feature)

    def delete_feature(self, name, namespace=None, delete_data=False):
        self._delete(model.Feature, namespace, name, delete_data=delete_data)

    def update_feature(self, name, namespace=None, **kwargs):
        self.__class__._validate_kwargs(
            kwargs,
            valid=["description", "meta", "transform"],
        )
        self._update(model.Feature, name=name, namespace=namespace, payload=kwargs)

    def transform(self, name, namespace=None, from_features=[]):
        def decorator(func):
            # Create or update feature with transform
            to_namespace, to_name = self._split_name(namespace=namespace, name=name)
            computed_from = [f"{ns}/{n}" for ns, n in self._unpack_list(from_features)]
            for feature in computed_from:
                assert self._exists(
                    model.Feature, name=feature
                ), f"{feature} does not exist in the feature store"

            transform = {"function": func, "args": computed_from}
            payload = {"transform": transform, "description": func.__doc__}
            if self._exists(model.Feature, namespace=to_namespace, name=to_name):
                # Already exists, update it
                self.update_feature(to_name, namespace=to_namespace, **payload)
            else:
                # Create a new feature
                self.create_feature(to_name, namespace=to_namespace, **payload)
            # Call the transform
            def wrapped_func(*args, **kwargs):
                return func(*args, **kwargs)

            return wrapped_func

        return decorator

    def load_dataframe(
        self,
        features,
        from_date=None,
        to_date=None,
        freq=None,
        time_travel=None,
    ):
        dfs = []
        # Load each requested feature
        for f in self._unpack_list(features):
            namespace, name = f
            with conn.session_scope(self.session_maker) as session:
                feature = (
                    session.query(model.Feature)
                    .filter_by(name=name, namespace=namespace)
                    .one_or_none()
                )
                if not feature:
                    raise MissingFeatureException(
                        f"No feature named {name} exists in {namespace}"
                    )
                # Load individual feature
                df = feature.load(
                    from_date=from_date,
                    to_date=to_date,
                    freq=freq,
                    time_travel=time_travel,
                )
                dfs.append(df.rename(columns={"value": f"{namespace}/{name}"}))
        return ts.concat(dfs)

    def save_dataframe(self, df, name=None, namespace=None):
        # Check dataframe columns
        feature_columns = df.columns.difference(["time", "created_time"])
        if len(feature_columns) == 1:
            # Single feature to save
            if feature_columns[0] == "value":
                if not name:
                    raise FeatureStoreException("Must specify feature name")
            else:
                name = feature_columns[0]
                df = df.rename(columns={name: "value"})
            if not self._exists(model.Feature, namespace, name):
                raise MissingFeatureException(
                    f"Feature named {name} does not exist in {namespace}"
                )
            # Save data for this feature
            namespace, name = self.__class__._split_name(namespace, name)
            with conn.session_scope(self.session_maker) as session:
                feature = (
                    session.query(model.Feature)
                    .filter_by(name=name, namespace=namespace)
                    .one()
                )
                # Save individual feature
                feature.save(df)
        else:
            # Multiple features in column names
            for feature_name in feature_columns:
                if not self._exists(model.Feature, namespace, name):
                    raise MissingFeatureException(
                        f"Feature named {name} does not exist in {namespace}"
                    )
            for feature_name in feature_columns:
                # Save individual features
                feature_df = df[[*df.columns.difference(feature_columns), feature_name]]
                self.save_dataframe(feature_df)

    def last(self, features):
        result = {}
        for f in self._unpack_list(features):
            namespace, name = f
            with conn.session_scope(self.session_maker) as session:
                feature = (
                    session.query(model.Feature)
                    .filter_by(name=name, namespace=namespace)
                    .one_or_none()
                )
                if not feature:
                    raise MissingFeatureException(
                        f"No feature named {name} exists in {namespace}"
                    )
                # Load individual feature
                result[f"{namespace}/{name}"] = feature.last()
        return result



from ._base import BaseStore
import posixpath
import dask.dataframe as dd
import dask
import pandas as pd
import numpy as np
import pyarrow as pa
import fsspec
import warnings


class Store(BaseStore):
    """Dask-backed timeseries data storage."""

    def __init__(self, url, storage_options={}):
        super().__init__(url, storage_options=storage_options)

    @staticmethod
    def _clean_dict(d):
        """Cleans dictionary of extraneous keywords."""
        remove_keys = ["_expires"]
        return {k: v for k, v in d.items() if k not in remove_keys}

    def _fs(self, name=None):
        fs, fs_token, paths = fsspec.get_fs_token_paths(
            self.url,
            storage_options=self._clean_dict(self.storage_options),
        )
        if name:
            feature_path = posixpath.join(paths[0], "feature", name)
        else:
            feature_path = posixpath.join(paths[0], "feature")
        return fs, feature_path

    def _full_feature_path(self, name):
        return posixpath.join(self.url, "feature", name)

    def _list_partitions(self, name, n=None, reverse=False):
        """List the available partitions for a feature."""
        fs, feature_path = self._fs(name)
        try:
            objects = fs.ls(feature_path)
        except FileNotFoundError:
            return []
        partitions = [obj for obj in objects if obj.startswith("partition=")]
        partitions = [p.split("=")[1] for p in partitions]
        partitions = sorted(partitions, reverse=reverse)
        if n:
            partitions = partitions[:n]
        return partitions

    @staticmethod
    def _apply_partition(partition, dt, offset=0):
        if isinstance(dt, dd.core.Series):
            if partition == "year":
                return dt.dt.year + offset
            elif partition == "date":
                return (dt + pd.Timedelta(days=offset)).dt.strftime("%Y-%m-%d")
            else:
                raise NotImplementedError(f"{partition} has not been implemented")

    def _write(self, name, ddf, **kwargs):
        # Write to output location
        feature_path = self._full_feature_path(name)
        # Build schema
        schema = {"time": pa.timestamp("ns"), "created_time": pa.timestamp("ns")}
        for field in pa.Table.from_pandas(ddf.head()).schema:
            if field.name in ["value", "partition"]:
                schema[field.name] = field.type
        try:
            ddf.to_parquet(
                feature_path,
                engine="pyarrow",
                compression="snappy",
                write_index=True,
                append=kwargs.get("append", False),
                partition_on="partition",
                ignore_divisions=True,
                schema=schema,
                storage_options=self._clean_dict(self.storage_options),
            )
        except Exception as e:
            raise RuntimeError(f"Unable to save data to {path}: {str(e)}")

    def _read(
        self, name, from_date=None, to_date=None, freq=None, time_travel=None, **kwargs
    ):
        # Identify which partitions to read
        filters = []
        if from_date:
            filters.append(("time", ">=", pd.Timestamp(from_date)))
        if to_date:
            filters.append(("time", "<=", pd.Timestamp(to_date)))
        if kwargs.get("partitions"):
            for p in kwargs.get("partitions"):
                filters.append(("partition", "==", p))
        filters = [filters] if filters else None
        # Read the data
        feature_path = self._full_feature_path(name)
        try:
            ddf = dd.read_parquet(
                feature_path,
                engine="pyarrow",
                filters=filters,
                storage_options=self._clean_dict(self.storage_options),
            )
            ddf = ddf.repartition(partition_size="25MB")
        except PermissionError as e:
            raise e
        except Exception as e:
            # No data available
            empty_df = pd.DataFrame(
                columns=["time", "created_time", "value", "partition"]
            ).set_index("time")
            ddf = dd.from_pandas(empty_df, chunksize=1)
        if "partition" in ddf.columns:
            ddf = ddf.drop(columns="partition")
        # Apply time-travel
        if time_travel:
            ddf = ddf.reset_index()
            ddf = ddf[ddf.created_time <= ddf.time + pd.Timedelta(time_travel)]
            ddf = ddf.set_index("time")
        # De-serialize from JSON if required
        if kwargs.get("serialized"):
            ddf = ddf.map_partitions(
                lambda df: df.assign(value=df.value.apply(pd.io.json.loads)),
                meta={
                    "value": "object",
                    "created_time": "datetime64[ns]",
                },
            )
        return ddf

    def ls(self):
        fs, path = self._fs()
        feature_names = [p.split("/")[-1] for p in fs.ls(path)]
        return feature_names

    def load(
        self, name, from_date=None, to_date=None, freq=None, time_travel=None, **kwargs
    ):
        # Find the last value _before_ time range to carry over
        last_before = from_date
        if from_date:
            _, last_before = self._range(
                name, to_date=from_date, time_travel=time_travel
            )
            last_before = last_before["time"]
        ddf = self._read(name, last_before, to_date, freq, time_travel, **kwargs)
        if not from_date:
            from_date = ddf.index.min().compute()  # First value in data
        if not to_date:
            to_date = ddf.index.max().compute()  # Last value in data
        if pd.Timestamp(to_date) < pd.Timestamp(from_date):
            to_date = from_date
        # Keep only last created_time for each index timestamp
        delayed_apply = dask.delayed(
            # Use pandas on each dask partition
            lambda x: x.reset_index()
            .set_index("created_time")
            .sort_index()
            .groupby("time")
            .last()
        )
        ddf = dd.from_delayed([delayed_apply(d) for d in ddf.to_delayed()])
        #  Repartition to remove empty chunks
        ddf = ddf.repartition(partition_size="25MB")
        # Apply resampling/date filtering
        if freq:
            # Index samples for final dataframe
            samples = dd.from_pandas(
                pd.DataFrame(index=pd.date_range(from_date, to_date, freq=freq)),
                chunksize=100000,
            )
            ddf = dd.merge(
                # Interpolate
                dd.merge(
                    ddf,
                    samples,
                    left_index=True,
                    right_index=True,
                    how="outer",
                ).ffill(),
                samples,
                left_index=True,
                right_index=True,
                how="right",
            )
        else:
            # Filter on date range
            ddf = ddf.loc[pd.Timestamp(from_date) : pd.Timestamp(to_date)]
        #  Repartition to remove empty chunks
        ddf = ddf.repartition(partition_size="25MB")
        return ddf

    def _range(self, name, **kwargs):
        ddf = self._read(name, **kwargs)
        # Don't warn when querying empty feature
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            first = ddf.head(1)
            last = ddf.tail(1)
        first = (
            {"time": None, "value": None}
            if first.empty
            else {"time": first.index[0], "value": first["value"].iloc[0]}
        )
        last = (
            {"time": None, "value": None}
            if last.empty
            else {"time": last.index[0], "value": last["value"].iloc[0]}
        )
        return first, last

    def first(self, name, **kwargs):
        first, _ = self._range(name, **kwargs)
        return first["value"]

    def last(self, name, **kwargs):
        _, last = self._range(name, **kwargs)
        return last["value"]

    def save(self, name, df, **kwargs):
        if df.empty:
            # Nothing to do
            return
        # Convert Pandas -> Dask
        if isinstance(df, pd.DataFrame):
            ddf = dd.from_pandas(df, chunksize=100000)
        elif isinstance(df, dd.DataFrame):
            ddf = df
        else:
            raise ValueError("Data must be supplied as a Pandas or Dask DataFrame")
        # Check value columm
        if "value" not in ddf.columns:
            raise ValueError("DataFrame must contain a value column")
        # Check we have a timestamp index column
        if np.issubdtype(ddf.index.dtype, np.datetime64):
            ddf = ddf.reset_index()
            if "time" in df.columns:
                raise ValueError(
                    "Not sure whether to use timestamp index or time column"
                )
        # Check time column
        partition = kwargs.get("partition", "date")
        if "time" in ddf.columns:
            ddf = ddf.assign(time=ddf.time.astype("datetime64[ns]"))
            # Add partition column
            ddf = ddf.assign(partition=self._apply_partition(partition, ddf.time))
            ddf = ddf.set_index("time")
        else:
            raise ValueError(
                f"DataFrame must be supplied with timestamps, not {ddf.index.dtype}"
            )
        # Check for created_time column
        if "created_time" not in ddf.columns:
            ddf = ddf.assign(created_time=pd.Timestamp.now())
        else:
            ddf = ddf.assign(created_time=ddf.created_time.astype("datetime64[ns]"))
        # Check for extraneous columns
        extraneous = set(ddf.columns) - set(["created_time", "value", "partition"])
        if len(extraneous) > 0:
            raise ValueError(f"DataFrame contains extraneous columns: {extraneous}")
        # Serialize to JSON if required
        if kwargs.get("serialized"):
            ddf = ddf.map_partitions(
                lambda df: df.assign(value=df.value.apply(pd.io.json.dumps))
            )
        # Save
        self._write(name, ddf, append=True)

    def delete(self, name):
        fs, feature_path = self._fs(name)
        try:
            fs.rm(feature_path, recursive=True)
        except FileNotFoundError:
            pass

    def _export(self, name):
        # Read the data
        feature_path = self._full_feature_path(name)
        try:
            ddf = dd.read_parquet(
                feature_path,
                engine="pyarrow",
                storage_options=self._clean_dict(self.storage_options),
            )
            # Repartition to optimise files on exported dataset
            ddf = ddf.repartition(partition_size="25MB")
            return ddf
        except Exception as e:
            # No data available
            return None

    def _import(self, name, ddf):
        if ddf is None or len(ddf.columns) == 0:
            return
        if "partition" not in ddf.columns:
            raise RuntimeError("Dask storage requires partitioning")
        # Copy data to new location
        self._write(name, ddf, append=False)



"""
Exception raised by Feature Store objects
"""


class FeatureStoreException(Exception):
    """General feature store exception."""

    pass


class RemoteFeatureStoreException(Exception):
    """General feature store exception."""

    pass


class MissingFeatureException(Exception):
    """The requested feature/namespace does not exist."""

    pass



import pandas as pd
import posixpath
from getpass import getpass
import time
import os
from pandas.io import json
from ._base import BaseFeatureStore
from . import _timeseries as ts
from . import _storage as storage
from . import _utils as utils
from .exceptions import *


try:
    # Allow for a minimal install with no requests
    import requests
    from requests_oauthlib import OAuth2Session
except ImportError:
    pass


class CloudFeatureStore(BaseFeatureStore):
    """**Cloud Feature Store**

    Connects to a hosted feature store via REST API.

    When using specifying features
    for `create_feature`, `update_feature`, etc., use either:

    * `namespace` and `name` as arguments; or
    * specify `name` in the format `"my-namespace/my-feature"`.
    """

    def __init__(
        self,
        connection_string="https://api.bytehub.ai",
        enable_transforms=True,
    ):
        """
        Args:
            connection_string (str): URL of ByteHub Cloud, e.g. [https://api.bytehub.ai](https://api.bytehub.ai).
            enable_transforms (bool, optional): whether to allow execution of pickled functions
                stored in the feature store. Required for feature transforms, but should only be enabled
                if you trust the feature store and the transforms that have been saved to it.
        """
        try:
            # Check that s3fs is available
            import s3fs
        except ImportError:
            raise RuntimeError(
                "Cloud feature store requires s3fs be installed: use pip install bytehub[cloud]"
            )
        self.enable_transforms = enable_transforms
        self._endpoint = connection_string
        if "/v1" not in self._endpoint:
            # Add API version
            self._endpoint = posixpath.join(self._endpoint, "v1/")
        if self._endpoint[-1] != "/":
            self._endpoint += "/"
        # Get the Oauth2 URLs
        response = requests.get(self._endpoint)
        self._check_response(response)
        self._urls = response.json()
        self._client_id = self._urls.pop("client_id")
        # Decide how to authenticate
        if os.environ.get("BYTEHUB_TOKEN"):
            # Non-interactive login using refresh token
            oauth = OAuth2Session(
                self._client_id,
                token={"refresh_token": os.environ.get("BYTEHUB_TOKEN")},
            )
            tokens = oauth.refresh_token(
                self._urls["token_url"],
                client_id=self._client_id,
                client_secret=None,
                include_client_id=True,
            )
        else:
            # Use interactive login
            oauth = OAuth2Session(
                self._client_id, redirect_uri=self._urls["callback_url"]
            )
            authorization_url, state = oauth.authorization_url(self._urls["login_url"])
            print(
                f"Please go to {authorization_url} and login. Copy the response code and paste below."
            )
            code_response = getpass("Response: ")
            tokens = oauth.fetch_token(
                self._urls["token_url"], code=code_response, include_client_id=True
            )
        self._tokens = tokens
        # Call list_namespaces to get temporary access tokens
        self.list_namespaces()

    def _check_response(self, response):
        # Raise an exception if response is not OK
        try:
            response.raise_for_status()
        except requests.exceptions.RequestException as e:
            try:
                payload = response.json()
                message = payload.get("message", payload.get("Message", ""))
            except requests.exceptions.RequestException:
                message = response.text
            raise RemoteFeatureStoreException(message)

    def _check_tokens(self):
        # Check that token hasn't expired
        if time.time() < self._tokens["expires_at"] - 10:
            return True
        else:
            # Token expired... refresh it
            if "refresh_token" in self._tokens:
                oauth = OAuth2Session(self._client_id, token=self._tokens)
                tokens = oauth.refresh_token(
                    self._urls["token_url"],
                    client_id=self._client_id,
                    client_secret=None,
                    include_client_id=True,
                )
                self._tokens = tokens
            else:
                raise RemoteFeatureStoreException("Cannot find refresh token")

    def _api_headers(self):
        # Headers to add to API requests
        self._check_tokens()
        return {
            "Authorization": self._tokens["access_token"],
            "Content-Type": "application/json",
        }

    def _refresh(self, name=None):
        # Check access token expiry, and refresh if required
        df = self._namespaces
        if name:
            df = df[df.name == name]
        for idx, row in df.iterrows():
            # Get storage options
            opt = row["storage_options"]
            if "_expires" in opt and pd.Timestamp(
                opt["_expires"]
            ) < pd.Timestamp.utcnow() + pd.Timedelta("1min"):
                self.list_namespaces()
                return

    def _exists(self, entity, **kwargs):
        if entity.lower() == "feature":
            ls = self.list_features(
                namespace=kwargs.get("namespace"), name=kwargs.get("name")
            )
        elif entity.lower() == "namespace":
            ls = self.list_namespaces(name=kwargs.get("name", kwargs.get("namespace")))
        else:
            raise MissingFeatureException(f"Unrecognised entity: {entity}")
        return not ls.empty

    def _get(self, entity, **kwargs):
        if entity.lower() == "feature":
            ls = self.list_features(
                namespace=kwargs.get("namespace"),
                name=kwargs.get("name"),
                friendly=False,
            )
        elif entity.lower() == "namespace":
            self._refresh(name=kwargs.get("name", kwargs.get("namespace")))
            ls = self._namespaces[
                self._namespaces.name == kwargs.get("name", kwargs.get("namespace"))
            ]
        else:
            raise MissingFeatureException(f"Unrecognised entity: {entity}")
        if len(ls) != 1:
            raise MissingFeatureException(f"{entity} not found: {kwargs}")
        return ls.iloc[0].to_dict()

    def _backend(self, namespace):
        ns = self._get("namespace", name=namespace)
        store = storage.available_backends["pandas"](
            url=ns["url"], storage_options=ns["storage_options"]
        )
        return store

    def list_namespaces(self, **kwargs):
        self.__class__._validate_kwargs(kwargs, ["name", "namespace", "regex"])
        url = self._endpoint + "namespace"
        response = requests.get(url, params=kwargs, headers=self._api_headers())
        self._check_response(response)
        df = pd.DataFrame(response.json())
        if df.empty:
            df = pd.DataFrame(
                columns=[
                    "namespace",
                    "name",
                    "version",
                    "description",
                    "meta",
                    "url",
                    "storage_options",
                ]
            )
        # Cache the namespace for use when loading/saving dataframes
        self._namespaces = df
        return df

    def create_namespace(self, name, **kwargs):
        self.__class__._validate_kwargs(
            kwargs,
            valid=["description", "url", "storage_options", "meta", "backend"],
            mandatory=["url"],
        )
        url = self._endpoint + "namespace"
        body = {"name": name, **kwargs}
        response = requests.post(
            url, data=json.dumps(body), headers=self._api_headers()
        )
        self._check_response(response)
        # Call list_namespaces to refresh namespace cache
        self.list_namespaces()

    def update_namespace(self, name, **kwargs):
        self.__class__._validate_kwargs(
            kwargs,
            valid=["description", "storage_options", "meta"],
        )
        url = self._endpoint + "namespace"
        body = {"name": name, **kwargs}
        response = requests.patch(
            url, data=json.dumps(body), headers=self._api_headers()
        )
        self._check_response(response)
        # Call list_namespaces to refresh namespace cache
        self.list_namespaces()

    def delete_namespace(self, name):
        if not self.list_features(namespace=name).empty:
            raise FeatureStoreException(
                f"{name} still contains features: these must be deleted first"
            )
        url = self._endpoint + "namespace"
        response = requests.delete(
            url, json={"name": name}, headers=self._api_headers()
        )
        self._check_response(response)
        # Call list_namespaces to refresh namespace cache
        self.list_namespaces()

    def clean_namespace(self, name):
        # Get namespace
        self._refresh(name=name)
        # Check for unused data and remove it
        store = self._backend(name)
        active_feature_names = self.list_features(namespace=name)
        active_feature_names = active_feature_names["name"].tolist()
        feature_data = store.ls()
        for feature in feature_data:
            if feature not in active_feature_names:
                # Redundant data... delete it
                store.delete(feature)

    def list_features(self, **kwargs):
        self.__class__._validate_kwargs(
            kwargs, valid=["name", "namespace", "regex", "friendly"]
        )
        url = self._endpoint + "feature"
        response = requests.get(url, params=kwargs, headers=self._api_headers())
        self._check_response(response)
        df = pd.DataFrame(response.json())
        if df.empty:
            return pd.DataFrame(
                columns=[
                    "namespace",
                    "name",
                    "version",
                    "description",
                    "meta",
                    "serialized",
                    "transform",
                    "partition",
                ]
            )
        if "transform" in df.columns and kwargs.get("friendly", True):
            df = df.assign(transform=df["transform"].apply(lambda x: x is not None))
        return df

    def create_feature(self, name, namespace=None, **kwargs):
        self.__class__._validate_kwargs(
            kwargs,
            valid=["description", "meta", "partition", "serialized", "transform"],
            mandatory=[],
        )
        url = self._endpoint + "feature"
        body = {"name": name, "namespace": namespace, **kwargs}
        response = requests.post(
            url, data=json.dumps(body), headers=self._api_headers()
        )
        self._check_response(response)

    def clone_feature(self, name, namespace=None, **kwargs):
        self.__class__._validate_kwargs(
            kwargs,
            valid=["from_namespace", "from_name"],
            mandatory=["from_name"],
        )
        from_namespace, from_name = self.__class__._split_name(
            kwargs.get("from_namespace"), kwargs.get("from_name")
        )
        to_namespace, to_name = self.__class__._split_name(namespace, name)
        if not self._exists("namespace", namespace=from_namespace):
            raise MissingFeatureException(f"{from_namespace} namespace does not exist")
        if not self._exists("namespace", namespace=to_namespace):
            raise MissingFeatureException(f"{to_namespace} namespace does not exist")
        # Get the existing feature
        payload = self._get("feature", namespace=from_namespace, name=from_name)
        _ = payload.pop("name")
        _ = payload.pop("namespace")
        _ = payload.pop("version")
        # Create the new feature
        self.create_feature(
            name=to_name,
            namespace=to_namespace,
            **payload,
        )
        # Copy data to new feature, if this raises exception will rollback
        if not payload.get("transform"):
            # Get storage backends for copy
            store_from = self._backend(from_namespace)
            store_to = self._backend(to_namespace)
            try:
                store_from.copy(from_name, to_name, store_to)
            except Exception as e:
                # Rollback... delete the newly created feature
                self.delete_feature(name=from_name, namespace=from_namespace)
                raise RemoteFeatureStoreException(
                    f"Unable to save data to {path}: {str(e)}"
                )

    def delete_feature(self, name, namespace=None, delete_data=False):
        url = self._endpoint + "feature"
        response = requests.delete(
            url,
            json={"name": name, "namespace": namespace},
            headers=self._api_headers(),
        )
        self._check_response(response)
        # Delete data
        if delete_data:
            namespace, name = self.__class__._split_name(namespace, name)
            store = self._backend(namespace)
            store.delete(name)

    def update_feature(self, name, namespace=None, **kwargs):
        self.__class__._validate_kwargs(
            kwargs,
            valid=["description", "meta", "transform"],
        )
        url = self._endpoint + "feature"
        body = {"name": name, "namespace": namespace, **kwargs}
        response = requests.patch(
            url, data=json.dumps(body), headers=self._api_headers()
        )
        self._check_response(response)

    def transform(self, name, namespace=None, from_features=[]):
        def decorator(func):
            # Create or update feature with transform
            to_namespace, to_name = self._split_name(namespace=namespace, name=name)
            computed_from = [f"{ns}/{n}" for ns, n in self._unpack_list(from_features)]
            existing_features = self.list_features()
            existing_features = existing_features.apply(
                lambda x: x["namespace"] + "/" + x["name"], axis=1
            ).tolist()
            for feature in computed_from:
                assert (
                    feature in existing_features
                ), f"{feature} does not exist in the feature store"

            transform = {
                "format": "cloudpickle",
                "function": utils.serialize(func),
                "args": computed_from,
            }
            payload = {"transform": transform, "description": func.__doc__}
            if self._exists("feature", namespace=to_namespace, name=to_name):
                # Already exists, update it
                self.update_feature(to_name, namespace=to_namespace, **payload)
            else:
                # Create a new feature
                self.create_feature(to_name, namespace=to_namespace, **payload)
            # Call the transform
            def wrapped_func(*args, **kwargs):
                return func(*args, **kwargs)

            return wrapped_func

        return decorator

    def _load_transform(
        self,
        feature,
        from_date,
        to_date,
        freq,
        time_travel,
        last=False,
        callers=[],
    ):
        # Check for recursive transforms
        full_name = feature["namespace"] + "/" + feature["name"]
        if full_name in callers:
            raise FeatureStoreException(
                f"Recursive feature transform detected on {full_name}"
            )
        # Load the transform function
        func = utils.deserialize(feature["transform"]["function"])
        # Load the features to transform
        dfs = []
        # Load each requested feature
        for f in feature["transform"]["args"]:
            namespace, name = f.split("/")[0], "/".join(f.split("/")[1:])
            f = self._get("feature", namespace=namespace, name=name)
            ns = self._get("namespace", name=namespace)
            # Load individual feature
            df = self._load(
                f,
                from_date,
                to_date,
                freq,
                time_travel,
                last=last,
                callers=[*callers, full_name],
            )
            dfs.append(df.rename(columns={"value": f"{namespace}/{name}"}))
        # Merge features into a single dataframe
        dfs = ts.concat(dfs)
        # Make sure columns are in the same order as args
        dfs = dfs[feature["transform"]["args"]]
        # Apply transform function
        transformed = ts.transform(dfs, func)
        return transformed

    def _load(
        self,
        feature,
        from_date,
        to_date,
        freq,
        time_travel,
        last=False,
        callers=[],
    ):
        if feature["transform"]:
            if not self.enable_transforms:
                raise RemoteFeatureStoreException(
                    "Transforms not enabled - if you trust this feature store then set enable_transforms=True when connecting"
                )
            # Apply feature transform
            return self._load_transform(
                feature,
                from_date,
                to_date,
                freq,
                time_travel,
                last,
                callers,
            )
        else:
            store = self._backend(feature["namespace"])
            # Change date range if getting last value
            if last:
                from_date = store.last(feature["name"])
                to_date = None
            return store.load(
                feature["name"],
                from_date=from_date,
                to_date=to_date,
                freq=freq,
                time_travel=time_travel,
                serialized=feature["serialized"],
            )

    def load_dataframe(
        self,
        features,
        from_date=None,
        to_date=None,
        freq=None,
        time_travel=None,
    ):
        dfs = []
        # Load each requested feature
        for f in self._unpack_list(features):
            namespace, name = f
            feature = self._get("feature", name=name, namespace=namespace)
            df = self._load(feature, from_date, to_date, freq, time_travel)
            dfs.append(df.rename(columns={"value": f"{namespace}/{name}"}))
        return ts.concat(dfs)

    def save_dataframe(self, df, name=None, namespace=None):
        # Check dataframe columns
        feature_columns = df.columns.difference(["time", "created_time"])
        if len(feature_columns) == 1:
            # Single feature to save
            if feature_columns[0] == "value":
                if not name:
                    raise FeatureStoreException("Must specify feature name")
            else:
                name = feature_columns[0]
                df = df.rename(columns={name: "value"})
            feature = self._get("feature", namespace=namespace, name=name)
            # Save data for this feature
            namespace, name = self.__class__._split_name(namespace, name)
            store = self._backend(namespace)
            # Save individual feature
            store.save(
                name,
                df,
                partition=feature["partition"],
                serialized=feature["serialized"],
            )
        else:
            # Multiple features in column names
            for feature_name in feature_columns:
                if not self._exists("feature", name=feature_name):
                    raise FeatureStoreException(
                        f"Feature named {name} does not exist in {namespace}"
                    )
            for feature_name in feature_columns:
                # Save individual features
                feature_df = df[[*df.columns.difference(feature_columns), feature_name]]
                self.save_dataframe(feature_df)

    def last(self, features):
        result = {}
        for f in self._unpack_list(features):
            namespace, name = f
            feature = self._get("feature", namespace=namespace, name=name)
            # Load individual feature
            df = self._load(
                feature,
                from_date=None,
                to_date=None,
                freq=None,
                time_travel=None,
                last=True,
            )
            r = df.tail(1)
            if r.empty:
                result[f"{namespace}/{name}"] = None
            else:
                result[f"{namespace}/{name}"] = r["value"].iloc[0]
        return result



from abc import ABC, abstractmethod
import functools
import types


class BaseStore(ABC):
    """Base storage class for ByteHub timeseries data.
    Sub-class this to add additional storage backends.
    """

    def __init__(self, url, storage_options={}):
        self.url = url
        self.storage_options = storage_options

    @abstractmethod
    def ls(self):
        """List features contained in storage."""
        raise NotADirectoryError()

    @abstractmethod
    def load(
        self, name, from_date=None, to_date=None, freq=None, time_travel=None, **kwargs
    ):
        """Load a single timeseries dataframe from storage."""
        raise NotImplementedError()

    @abstractmethod
    def save(self, name, df, **kwargs):
        """Save a timeseries dataframe."""
        raise NotImplementedError()

    @abstractmethod
    def first(self, name, **kwargs):
        """Retrieves first index from the timeseries."""
        raise NotImplementedError()

    def last(self, name, **kwargs):
        """Retrieves last index from the timeseries."""
        raise NotImplementedError()

    @abstractmethod
    def delete(self, name):
        """Delete the data for a feature."""
        raise NotImplementedError()

    def copy(self, from_name, to_name, destination_store):
        """Used during clone operations to copy timeseries data between locations.
        Override this to implement more efficient copying between specific storage backends.
        """
        # Export existing data
        ddf = self._export(from_name)
        # Import to destination
        if isinstance(ddf, types.GeneratorType):
            for ddf_ in ddf:
                destination_store._import(to_name, ddf_)
        else:
            destination_store._import(to_name, ddf)

    @abstractmethod
    def _export(self, name):
        """Export a timeseries as standardised dataframe."""
        raise NotImplementedError()

    @abstractmethod
    def _import(self, name, ddf):
        """Import a timeseries from standardised dataframe."""
        raise NotImplementedError()



from setuptools import setup, find_packages
from os import path


this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_desc = f.read()

with open(path.join(this_directory, "requirements.txt"), encoding="utf-8") as f:
    install_requires = f.read()

exec(open("bytehub/_version.py").read())

setup(
    name="bytehub",
    packages=find_packages(),
    version=__version__,
    description="ByteHub Timeseries Feature Store",
    author="ByteHub AI Limited",
    url="https://bytehub.ai",
    long_description=long_desc,
    long_description_content_type="text/markdown",
    install_requires=install_requires,
    extras_require={
        "dev": ["pytest", "black", "bump2version", "jupyterlab", "pre-commit", "pdoc3"],
        "aws": ["s3fs==0.4"],
        "gcp": ["gcsfs==0.8"],
        "azure": ["adlfs>=0.6"],
        "cloud": ["s3fs==0.4"],
    },
)



from .dask import Store as Dask
import pandas as pd


class Store(Dask):
    """Pandas-backed timeseries data storage (uses Dask for parquet functions)."""

    def __init__(self, url, storage_options={}):
        super().__init__(url, storage_options=storage_options)

    def load(
        self, name, from_date=None, to_date=None, freq=None, time_travel=None, **kwargs
    ):
        # Find the last value _before_ time range to carry over
        last_before = from_date
        if from_date:
            _, last_before = self._range(
                name, to_date=from_date, time_travel=time_travel
            )
            last_before = last_before["time"]
        ddf = self._read(name, last_before, to_date, freq, time_travel, **kwargs)
        if not from_date:
            from_date = ddf.index.min().compute()  # First value in data
        if not to_date:
            to_date = ddf.index.max().compute()  # Last value in data
        if pd.Timestamp(to_date) < pd.Timestamp(from_date):
            to_date = from_date
        pdf = ddf.compute()
        # Keep only last created_time for each index timestamp
        pdf = (
            pdf.reset_index()
            .set_index("created_time")
            .sort_index()
            .groupby("time")
            .last()
        )
        # Apply resampling/date filtering
        if freq:
            samples = pd.DataFrame(index=pd.date_range(from_date, to_date, freq=freq))
            pdf = pd.merge(
                pd.merge(
                    pdf, samples, left_index=True, right_index=True, how="outer"
                ).ffill(),
                samples,
                left_index=True,
                right_index=True,
                how="right",
            )
        else:
            # Filter on date range
            pdf = pdf.loc[pd.Timestamp(from_date) : pd.Timestamp(to_date)]
        return pdf



import json
import base64
import cloudpickle


def is_jsonable(x):
    try:
        json.dumps(x)
        return True
    except:
        return False


def serialize(func):
    return base64.b64encode(cloudpickle.dumps(func)).decode("utf-8")


def deserialize(func_string):
    return cloudpickle.loads(base64.b64decode(func_string.encode("utf-8")))



from abc import ABC
import pandas as pd


class BaseFeatureStore(ABC):
    """Base class for ByteHub feature stores."""

    @classmethod
    def _split_name(cls, namespace=None, name=None):
        """Parse namespace and name."""
        if not namespace and name and "/" in name:
            parts = name.split("/")
            namespace, name = parts[0], "/".join(parts[1:])
        return namespace, name

    @classmethod
    def _validate_kwargs(cls, args, valid=[], mandatory=[]):
        for name in args.keys():
            if name not in valid:
                raise ValueError(f"Invalid argument: {name}")
        for name in mandatory:
            if name not in args.keys():
                raise ValueError(f"Missing mandatory argument: {name}")

    @classmethod
    def _unpack_list(cls, obj, namespace=None):
        """Extract namespace, name combinations from DataFrame or list
        and return as list of tuples
        """
        if isinstance(obj, str):
            return [cls._split_name(namespace=namespace, name=obj)]
        elif isinstance(obj, pd.DataFrame):
            # DataFrame format must have a name column
            df = obj
            if "name" not in df.columns:
                raise ValueError("DataFrame must have a name column")
            return [
                (row.get("namespace", namespace), row.get("name"))
                for _, row in df.iterrows()
            ]
        elif isinstance(obj, list):
            # Could be list of names, of list of dictionaries
            r = []
            for item in obj:
                if isinstance(item, str):
                    r.append(cls._split_name(name=item, namespace=namespace))
                elif isinstance(item, dict):
                    r.append(
                        cls._split_name(
                            namespace=item.get("namespace"), name=item.get("name")
                        )
                    )
                else:
                    raise ValueError("List must contain strings or dicts")
            return r
        else:
            raise ValueError(
                "Must supply a string, dataframe or list specifying namespace/name"
            )

    def __init__(self, connection_string="sqlite:///bytehub.db", backend="pandas"):
        """Create a Feature Store, or connect to an existing one."""
        raise NotImplementedError()

    def list_namespaces(self, **kwargs):
        """List namespaces in the feature store.

        Search by name or regex query.

        Args:
            name (str, optional): name of namespace to filter by.
            namespace (str, optional): same as name.
            regex (str, optional): regex filter on name.

        Returns:
            pd.DataFrame: DataFrame of namespaces and metadata.
        """
        raise NotImplementedError()

    def create_namespace(self, name, **kwargs):
        """Create a new namespace in the feature store.

        Args:
            name (str): name of the namespace.
            description (str, optional): description for this namespace.
            url (str): url of data store.
            storage_options (dict, optional): storage options, e.g. access credentials.
            backend (str, optional): storage backend, see bytehub._storage.available_backends, defaults to `"pandas"`.
            meta (dict, optional): key/value pairs of metadata.
        """
        raise NotImplementedError()

    def update_namespace(self, name, **kwargs):
        """Update a namespace in the feature store.

        Args:
            name (str): namespace to update.
            description (str, optional): updated description.
            storage_options (dict, optional): updated storage options.
            meta (dict, optional): updated key/value pairs of metadata.
                To remove metadata, update using `{"key_to_remove": None}`.
        """
        raise NotImplementedError()

    def delete_namespace(self, name):
        """Delete a namespace from the feature store.

        Args:
            name: namespace to be deleted.
        """
        raise NotImplementedError()

    def clean_namespace(self, name):
        """Removes any data that is not associated with features in the namespace.
        Run this to free up disk space after deleting features

        Args:
            name (str): namespace to clean
        """
        raise NotImplementedError()

    def list_features(self, **kwargs):
        """List features in the feature store.

        Search by namespace, name and/or regex query

        Args:
            name (str, optional): name of feature to filter by.
            namespace (str, optional): namespace to filter by.
            regex (str, optional): regex filter on name.
            friendly (bool, optional): simplify output for user.

        Returns:
            pd.DataFrame: DataFrame of features and metadata.
        """
        raise NotImplementedError()

    def create_feature(self, name, namespace=None, **kwargs):
        """Create a new feature in the feature store.

        Args:
            name (str): name of the feature
            namespace (str, optional): namespace which should hold this feature.
            description (str, optional): description for this namespace.
            partition (str, optional): partitioning of stored timeseries (default: `"date"`).
            serialized (bool, optional): if `True`, converts values to JSON strings before saving,
                which can help in situations where the format/schema of the data changes
                over time.
            transform (str, optional): pickled function code for feature transforms.
            meta (dict, optional): key/value pairs of metadata.
        """
        raise NotImplementedError()

    def clone_feature(self, name, namespace=None, **kwargs):
        """Create a new feature by cloning an existing one.

        Args:
            name (str): name of the feature.
            namespace (str, optional): namespace which should hold this feature.
            from_name (str): the name of the existing feature to copy from.
            from_namespace (str, optional): namespace of the existing feature.
        """
        raise NotImplementedError()

    def delete_feature(self, name, namespace=None, delete_data=False):
        """Delete a feature from the feature store.

        Args:
            name (str): name of feature to delete.
            namespace (str, optional): namespace, if not included in feature name.
            delete_data (bool, optional): if set to `True` will delete underlying stored data
                for this feature, otherwise default behaviour is to delete the feature store
                metadata but leave the stored timeseries values intact.
        """
        raise NotImplementedError()

    def update_feature(self, name, namespace=None, **kwargs):
        """Update a feature in the feature store.

        Args:
            name (str): feature to update.
            namespace (str, optional): namespace, if not included in feature name.
            description (str, optional): updated description.
            transform (str, optional): pickled function code for feature transforms.
            meta (dict, optional): updated key/value pairs of metadata.
                To remove metadata, update using `{"key_to_remove": None}`.
        """
        raise NotImplementedError()

    def transform(self, name, namespace=None, from_features=[]):
        """Decorator for creating/updating virtual (transformed) features.
        Use this on a function that accepts a dataframe input and returns an output dataframe
        of tranformed values.

        Args:
            name (str): feature to update.
            namespace (str, optional): namespace, if not included in feature name.
            from_features (list): list of features which should be transformed by this one
        """
        raise NotImplementedError()

    def load_dataframe(
        self,
        features,
        from_date=None,
        to_date=None,
        freq=None,
        time_travel=None,
    ):
        """Load a DataFrame of feature values from the feature store.

        Args:
            features (Union[str, list, pd.DataFrame]): name of feature to load, or list/DataFrame of feature namespaces/name.
            from_date (datetime, optional): start date to load timeseries from, defaults to everything.
            to_date (datetime, optional): end date to load timeseries to, defaults to everything.
            freq (str, optional): frequency interval at which feature values should be sampled.
            time_travel (str, optional): timedelta string, indicating that time-travel should be applied to the
                returned timeseries values, useful in forecasting applications.

        Returns:
            Union[pd.DataFrame, dask.DataFrame]: depending on which backend was specified in the feature store.
        """
        raise NotImplementedError()

    def save_dataframe(self, df, name=None, namespace=None):
        """Save a DataFrame of feature values to the feature store.

        Args:
            df (pd.DataFrame): DataFrame of feature values.
                Must have a `time` column or DateTimeIndex of time values.
                Optionally include a `created_time` column (defaults to `utcnow()` if omitted).
                For a single feature: a `value` column, or column header of feature `namespace/name`.
                For multiple features name the columns using `namespace/name`.
            name (str, optional): name of feature, if not included in DataFrame column name.
            namespace (str, optional): namespace, if not included in DataFrame column name.
        """
        raise NotImplementedError()

    def last(self, features):
        """Fetch the last value of one or more features.

        Args:
            features (Union[str, list, pd.DataFrame]): feature or features to fetch.

        Returns:
            dict: dictionary of name, last value pairs.
        """
        raise NotImplementedError()

    def create_task(self):
        """Create a scheduled task to update the feature store."""
        raise NotImplementedError()

    def update_task(self):
        """Update a task."""
        raise NotImplementedError()

    def delete_task(self):
        """Delete a task."""
        raise NotImplementedError()
